#include "../../includes/MeasurementSystem.h"
#include <stdlib.h>

//Internal storage variables
sensorMeasurement measurementBuffer[MEASUREMENT_BUFFER_SIZE];
Sensor *sensorArray[SENSOR_COUNT];
int measurementCount = 0;
void _addMSRtoArray(sensorMeasurement *msr);

void _addMSRtoArray(sensorMeasurement *msr){
    if (measurementCount == 150){//Buffer is full, overwrite the previous measurement at index 149
        measurementBuffer[149] = *msr;
    }
    else{
        measurementBuffer[measurementCount] = *msr;
        measurementCount += 1;
    }
}

void MeasurementInit(){
    static cJSON_Hooks hooks = {
        .malloc_fn = k_malloc,
        .free_fn = k_free,
    };
    cJSON_InitHooks(&hooks);
    sensorArray[0] = &sensor_VBatt;
    sensorArray[1] = &sensor_VSys;
}

//Returns number of measurements currently in the buffer based upon the measurementCount
int MeasurementgetCountofMeasuresinBuffer(){return measurementCount;}

//Asks a given sensor to perform a measurement
int MeasurementmeasureSensor(int sensorID){
    if(sensorArray[sensorID]->powered == false){
        sensorArray[sensorID]->powerOn();
    }
    sensorMeasurement msrBuffer;
    int rslt = sensorArray[sensorID]->performMeasurement(&msrBuffer);
    _addMSRtoArray(&msrBuffer);
    sensorArray[sensorID]->timeLastMeasured = msrBuffer.timeTaken;
    return rslt;
}

sensorMeasurement* MeasurementgetLastMeasurement(){
    return &measurementBuffer[measurementCount-1];
}


uint64_t MeasurementgetTimeofNextMeasurement(){
    uint64_t nearestTime = __UINT64_MAX__;
    for (int i = 0; i < SENSOR_COUNT; i++){
        uint64_t timeNext = sensorArray[i]->timeLastMeasured + sensorArray[i]->updateInterval;
        if (timeNext < nearestTime) { nearestTime = timeNext;}
    }
    return nearestTime;
}

void MeasurementMeasureExpired(){
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    LOG_INF("Starting measurement of sensors");
    uint64_t timeStamp;
    RTCgetTime(&timeStamp);
    for (int i = 0; i < SENSOR_COUNT; i++){
        if ((sensorArray[i]->timeLastMeasured + sensorArray[i]->updateInterval) < (timeStamp + MEASUREMENT_TIME_WIGGLE)){ 
            MeasurementmeasureSensor(sensorArray[i]->sensorID);
        }
    }
    LOG_INF("Completed, current measurement count: %d ", measurementCount);
}

//Function export measurements as a JSON
//This needs to be freed after being used
char *dumpMeasurementsToJson(){ 
    //Set up the JSON Object
    cJSON *data = NULL;
    cJSON *msrs = NULL;
    data = cJSON_CreateObject();
    cJSON_AddStringToObject(data, "System_Name", BOARD_IDENTIFIER);
    uint64_t timeBuffer;
    RTCgetTime(&timeBuffer);
    cJSON_AddNumberToObject(data, "Generated At", (double) timeBuffer);
    msrs = cJSON_CreateArray();
    for (int i = 0; i < measurementCount; i ++){
        cJSON *msr = NULL;
        msr = cJSON_CreateObject();
        cJSON_AddNumberToObject(msr, "ID", (double) measurementBuffer[i].id);
        char msr_type_buffer[2] = {measurementBuffer[i].type, "\0"}; //Char to stirng promotion
        cJSON_AddStringToObject(msr, "TYPE", msr_type_buffer);
        cJSON_AddNumberToObject(msr, "PRECISION", measurementBuffer[i].precision);
        cJSON_AddNumberToObject(msr, "VALUE", (double) measurementBuffer[i].value);
        cJSON_AddNumberToObject(msr, "TIMETAKEN", (double) measurementBuffer[i].timeTaken);
        cJSON_AddItemToArray(msrs, msr);
    }
    cJSON_AddItemToObject(data, "Measurements", msrs);
    //Wrap things up
    measurementCount = 0;
    char* output = cJSON_Print(data);
    cJSON_Delete(data);
    return output;
}

char* LatestToJson(void){
    cJSON *value = NULL;
    char * jsonOutput = NULL;
    value = cJSON_CreateObject();
    for(int i = measurementCount-1; i > (measurementCount-1)-SENSOR_COUNT; i--){
        //QUICK hack for names! NEEDS TO BE CHANGED
        char name[20] = {};
        if (measurementBuffer[i].id == 0){
            cJSON_AddNumberToObject(value, "Battery-Voltage", (double) measurementBuffer[i].value);
        }
        else{
            cJSON_AddNumberToObject(value, "System-Voltage", (double) measurementBuffer[i].value);
        }
    }
    jsonOutput = cJSON_Print(value);
    cJSON_Delete(value);
    return jsonOutput;
}

//Function for testing JSON export
int __shell_Dump_Measurements_to_Json(const struct shell *sh, size_t argc, char **argv){
    char *Json = NULL;
    Json = dumpMeasurementsToJson();
    shell_print(sh, "%s", Json);
    free(Json);
    return 0;
}
int __shell_Measure_Primary_Board_Sensors(const struct shell *sh, size_t argc, char **argv){
    for (int i = 0; i < SENSOR_COUNT; i++){
        MeasurementmeasureSensor(sensorArray[i]->sensorID);
    }
    return 0;
}

int __shell_Measure_get_Count(const struct shell *sh, size_t argc, char **argv){
    int temp = MeasurementgetCountofMeasuresinBuffer();
    shell_print(sh, "%d", temp);
}
