#include "../includes/timerInteruptHandler.h"

struct k_work _interuptWorkItem;
struct k_work _timeUpdateWorkItem;

uint64_t _MEASUREMENT_TimeLast_Uploaded = 0;

//Internal Functions forward decleration
uint64_t _getTimeofNextEvent();

void interuptHandlerInit(){
    k_work_init(&_interuptWorkItem, interuptTask);
    uint64_t tempTime;
    RTCgetTime(&tempTime);
    _MEASUREMENT_TimeLast_Uploaded = tempTime; //This prevents the firmware from trying to immediatly trying to upload code
    interuptTask();
}

void timer_alarm_isr(const struct device *dev, struct gpio_callback *cb, uint32_t pins){
    k_work_submit(&_interuptWorkItem);  
}

void interuptTask(){
    LOG_MODULE_DECLARE(firmware_log, LOG_LEVEL_INF);
    LOG_INF("Interupt triggered, performing periodic updates");
    uint64_t timeBuffer;
    RTCgetTime(&timeBuffer);
    //Perform any period tasks that need to be checked for
    MeasurementMeasureExpired();
    //Check if measurements should be uploaded to the cloud and then do so if needed
    if(MeasurementgetCountofMeasuresinBuffer() > MEASURMENT_MAX_BUFFER_FILL_SIZE
            || timeBuffer > (MEASUREMENT_MAX_TIME_BETWEEN_CLOUD_UPDATES + _MEASUREMENT_TimeLast_Uploaded))
    {
        if(!NetworkState->networkActive){ //Ensure the network is active
            int networkStatus = Networkconnect();
            if(!NetworkState->azureConnected && networkStatus == 0){
                Network_initCloud();
            }
            else if (networkStatus != 0) //Case in which it wasn't possible to connect to the network for some reason
            {
                LOG_WRN("Unable to connect to the network!");
                return; //This means that after the next measurement we will try agian.
                /*
                    This could be expanded in the future to include some form of better storage or something like that. 
                    As it currently stands measurements will continue to pile into the buffere until it starts activating the
                    circular buffer functionality and data is lost. This could of course be offset by simply expanding the size
                    of the buffer of course.
                */
            }
            
        }
        if(!NetworkState->azureConnected){
            Network_ConnectToCloud();
        }
        char *measurementJson = NULL;
        char *latestJson = NULL;
        latestJson = LatestToJson();
        measurementJson = dumpMeasurementsToJson(); //TODO: Add in some kind of handling here for the case where the network can't be reached or something like that
        Network_MeasurementToCloud(measurementJson); 
        Network_sendTo_Thingsboard(latestJson);
        k_free(measurementJson);
        k_free(latestJson);
        //Since the network is already connected it doesn't make sense not to pull down the time
        NetworksyncRTCClockSNTP();
        RTCgetTime(&timeBuffer);
        _MEASUREMENT_TimeLast_Uploaded = timeBuffer;
    }
    //Other tasks to do here:
    //  -Check to see if there are any firmware updates that need to be applied to the device
    //  -Check the data to see if we need to go into low power mode
    //      -> Note that this needs to only happen if an option for the installation is set (makes no sense in the subway)
    //Set the timer to interupt the MCU again when the next measurement needs to be taken
    timeBuffer = _getTimeofNextEvent();
    uint64_t curTime;
    RTCgetTime(&curTime);
    if (timeBuffer < curTime){ //Ensure we never lock yourselfs waiting for a time that has already passed
        timeBuffer = curTime + 60;
    }
    RTCsetINTAlarm(&timeBuffer);
}

uint64_t _getTimeofNextEvent(){
    uint64_t tempTime = MeasurementgetTimeofNextMeasurement();
    tempTime = (tempTime < (_MEASUREMENT_TimeLast_Uploaded+MEASUREMENT_MAX_TIME_BETWEEN_CLOUD_UPDATES)) ? tempTime : (_MEASUREMENT_TimeLast_Uploaded + MEASUREMENT_MAX_TIME_BETWEEN_CLOUD_UPDATES);
    return tempTime; 
}